/* Copyright 1997 Sun Microsystems, Inc.  All rights reserved. */

#ifndef shade_h_
#define shade_h_

#pragma ident "@(#)Shade and Spixtools V5.32C, V9 SPARC SunOS"


#include <spixtypes.h>


typedef union {
	int32	ii[2];
	uint32	uu[2];
#ifdef INT64
	int64	x;
#endif
} xreg_t;


typedef struct {
	xreg_t	sh_r[32];	/* integer registers */
	int32	sh_y;		/* y register */
	int8	sh_icc;		/* integer condition codes (see below) */
	int8	sh_xcc;		/* extended integer condition codes */
	uint8	sh_asi;		/* address space identifier */
	uint8	sh_gsr;		/* graphic status register */

	int32	sh_fr[128];	/* floating point registers */
	uint32	sh_fsr;		/* floating point state register, lsw */
	uint32	sh_fcc[3];	/* fp condition codes 1-3 (in %fsr format) */
} Shade;


#define sh_g0	sh_r[0]
#define sh_g1	sh_r[1]
#define sh_g2	sh_r[2]
#define sh_g3	sh_r[3]
#define sh_g4	sh_r[4]
#define sh_g5	sh_r[5]
#define sh_g6	sh_r[6]
#define sh_g7	sh_r[7]

#define sh_o0	sh_r[8]
#define sh_o1	sh_r[9]
#define sh_o2	sh_r[10]
#define sh_o3	sh_r[11]
#define sh_o4	sh_r[12]
#define sh_o5	sh_r[13]
#define sh_o6	sh_r[14]
#define sh_o7	sh_r[15]

#define sh_l0	sh_r[16]
#define sh_l1	sh_r[17]
#define sh_l2	sh_r[18]
#define sh_l3	sh_r[19]
#define sh_l4	sh_r[20]
#define sh_l5	sh_r[21]
#define sh_l6	sh_r[22]
#define sh_l7	sh_r[23]

#define sh_i0	sh_r[24]
#define sh_i1	sh_r[25]
#define sh_i2	sh_r[26]
#define sh_i3	sh_r[27]
#define sh_i4	sh_r[28]
#define sh_i5	sh_r[29]
#define sh_i6	sh_r[30]
#define sh_i7	sh_r[31]

#define sh_fp	sh_i6
#define sh_sp	sh_o6


/*
 * sh_icc and sh_xcc component values:
 */

#define SH_ICC_SHIFT	3
#define SH_ICC_N	(8 << SH_ICC_SHIFT)	/* negative */
#define SH_ICC_Z	(4 << SH_ICC_SHIFT)	/* zero */
#define SH_ICC_V	(2 << SH_ICC_SHIFT)	/* overflow */
#define SH_ICC_C	(1 << SH_ICC_SHIFT)	/* carry */

#define SH_CCR(sh)	((sh)->sh_icc >> SH_ICC_SHIFT & 15 | \
			((sh)->sh_xcc >> SH_ICC_SHIFT & 15) << 4)


/*
 * Possible status codes for shade_appstatus.
 */
typedef enum {
	SHADE_NOAPP,	/* no application loaded */
	SHADE_RUNNING,	/* application running */
	SHADE_EXITED,	/* application terminated due to exit() call */
	SHADE_ERRORED	/* application terminated due to error */
} shade_status_t;


#ifdef __STDC__

extern shade_status_t 	shade_appstatus(int *);

#else

extern shade_status_t 	shade_appstatus();

#endif /*__STDC__*/


#endif	/* #ifndef shade_h_ */
