#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <limits.h>
#include <signal.h>

#ifndef SIGUSR1
#define SIGUSR1 30
#endif

#define addr_t foo
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#include <trace.h>
#include <stdtr.h>
#include <trctl.h>
#include <IHASH.h>
#include <ITYPES.h>

#include "SimParams.h"
#include "my_socket.h"
#include "address_list.h"

#undef addr_t

#ifndef PATH_MAX
#define PATH_MAX _POSIX_PATH_MAX
#endif

#define alloc(n,t)	((t *) domalloc ((unsigned int)((n) * sizeof (t))))
#define percent(a,b)	(100. * (a) / (b))

#define align(a,b)	(((a) >> (b)) << (b))
#define sizemask(x)	((1 << (x)) - 1)

#define MAXSIZE		32	/* log2 of max cache reference operand size */

#define NLEVELS		2	/* max # of cache levels */
#define R		0	/* read */
#define W		1	/* write */
#define	AT_MORE		5

typedef uint32	addr_t;
typedef uint32	mask_t;


#define C_LRU		1	/* least recently used set replacement */
#define C_RANDOM	2	/* random set replacement */
#define C_VIKING	3	/* viking set replacement */

typedef struct CACHE 
{
	double	c_nref[1+MAXSIZE][2];
				/* # of accesses [size][read,write] */
	double	c_nmiss[1+MAXSIZE][2];
				/* # of misses [size][read,write] */
	double	c_nwback;	/* # of write backs from this cache */
	double	c_nrw;		/* # 1st write hit after read (write back) */
	double	c_nbkinving;	/* # subblocks victimized -> back invalidates */
	double	c_nbkinved;	/* # subblocks back invalidated */
	double	c_nwinv;	/* # subblocks write invalidated */
	addr_t	**c_tags;	/* address tags [#sets][#lines] */
	mask_t	**c_valid;	/* (subblock) valid bit(s) [#sets][#lines] */
	mask_t	**c_dirty;	/* (subblock) modified bit(s) [#sets][#lines] */
	char	**c_used;	/* block used [#sets][#lines] */
	mask_t	*c_mru;		/* mru bits for viking [#lines] */
	uint32	**c_lru;	/* lru set replacement counters [1+#sets][#lines] */
	int	c_nlines;	/* # cache lines */
	short	c_bsize;	/* log2 of cache block size (bytes) */
	short	c_bbsize;	/* log2 of cache subblock size (bytes) */
	short	c_ssize;	/* log2 of set size (bytes) */
	short	c_nsets;	/* # of sets; 1 for direct mapped */
	short	c_random;	/* "random" set replacement pointer */
	char	c_replace;	/* set replacement: C_RANDOM, etc. */
	char	c_wthru;	/* write through (else write back) */
	char	c_walloc;	/* write allocate */
	char	c_winv;		/* write invalidate (wt+nowa only) */
	char	c_idu;		/* 'I' for instrucion cache, etc. */
	char	c_level;	/* level # in cache hierarchy */
	char	c_isbkinv;	/* on some other caches bkinv list */
	int		c_sim;		/* simultaneous simulation number ... new ... */
	struct CACHE **c_bkinv;	/* caches to back invalidate (inclusion) */
	struct CACHE *c_next;	/* next cache level */
} Cache;

int	HOW_MANY = 0;
int	READ_COMMANDS_FROM_FILE = 0;
int	array_L1_I_Cache_miss[AT_MORE];
int	array_L1_D_Cache_miss[AT_MORE];
int	array_L2_Cache_miss[AT_MORE];

int	array_L1_I_Total_References[AT_MORE];
int	array_L1_D_Total_References[AT_MORE];
int	array_L2_Total_References[AT_MORE];

int	REPEATS;

extern 	FILE	*statsfp;
extern 	char	ihash_memsize[];
extern 	double	nina;
extern	int	START_SIMULATION;
extern 	void	exit();
extern	void	DtbPrintResults();
extern	long	DtbReference();
extern	long	InstrDtbReference();
extern	int	FD[2];

extern	int     InstrCacheSize;
extern	int     DataCacheSize;
extern	int     InstrTlbSize;
extern	int     DataTlbSize;
extern	int     SecondLeveCacheSize;

extern	int	DTBSIZE[5];
extern	int	InstrDTBSIZE[5];

extern	void	init_Tlb();

extern	int	milolid_main();
extern	addr_list *search_addr_list();

void		init_my_statistic_arrays();
void		PrintUsefulResults();

char		*anal_usage = "-{idu}<sz>[b<bsz>[,<subbsz>]][s<assoc>][r<replace>][wb][wt][wa][wi][I<cache>]...",
		*anal_version = "5.3 (24/Aug/94)";

extern long	strtol();

static Cache	icaches[NLEVELS],	/* instruction caches */
		dcaches[NLEVELS];	/* data caches */
static Cache    ficache[NLEVELS*AT_MORE]; /*  new */
static Cache    fdcache[NLEVELS*AT_MORE];     /* new */


static addr_t	getsize();

static char	*domalloc(),
		*ordinal(),
		*putsize();

static int	nlevels,	/* # of cache levels */
		replacecacheset();

static void	accesscache(),
		accesscacheset(),
		backinvalidatedcache(),
		backinvalidatingcache(),
		backinvalidatingcachesubblock(),
		cleancache(),
		cleancaches(),
		cleancacheline(),
		invalidatecache(),
		newcache(),
		newcachesets(),
		resetcache(),
		resetcaches(),
		resetcachesets();

int		L1_I_Cache_miss = 0;
int		L1_D_Cache_miss = 0;
int		L2_Cache_miss = 0;

int		L1_I_Total_References = 0;
int		L1_D_Total_References = 0;
int		L2_Total_References = 0;

Msg_Info	Update_Interface_Info;
Msg_simultaneous	Update_Simultaneous_Interface;

extern	int	*array_misses, *array_hits, *array_InstrMisses, *array_InstrHits;
extern	int	connection;

int		TOTAL_COUNTER[AT_MORE+1];
int		CONNECTION_IS_OK = 0;

int ex_misses = 0, ex_hits = 0, ex_InstrMisses = 0, ex_InstrHits = 0;
int array_ex_misses[6], array_ex_hits[6], array_ex_InstrMisses[6], array_ex_InstrHits[6]; /* 6: AT_MORE + 1 ... */

int ex_L1_D_Cache_miss = 0, ex_L1_D_Total_References = 0;
int array_ex_L1_D_Cache_miss[5], array_ex_L1_D_Total_References[5];

int ex_L1_I_Cache_miss = 0, ex_L1_I_Total_References = 0;
int array_ex_L1_I_Cache_miss[5], array_ex_L1_I_Total_References[5];

int ex_L2_Cache_miss = 0, ex_L2_Total_References = 0;	
int array_ex_L2_Cache_miss[5], array_ex_L2_Total_References[5];

FILE		*abel_file;

void
init_information_arrays()
{
	int	i;

	for (i=0; i < 5; i++)
	{
		InstrTLBParams[i] = 8;
		DataTLBParams[i] = 8;
		InstrCacheParams[i] = 5;
		DataCacheParams[i] = 128;
		SecLevelCacheParams[i] = 256;
		
		strcpy(InstrTLBWhat[i],"Positions");
		strcpy(DataTLBWhat[i],"Positions");
		strcpy(InstrCacheWhat[i],"Kb");
		strcpy(DataCacheWhat[i],"Kb");
		strcpy(SecLevelCacheWhat[i],"Kb");
	}
}

void
initialize ()
{
	Cache	*c;
	int	a, i, id = 0;
	int	j, k;	/* new */
	static	int	first_time_only = 0;

	nlevels = 0;

	for (i = 0; i < NLEVELS; i++)
		icaches[i].c_level = dcaches[i].c_level = i + 1;

	for (j = 0; j < HOW_MANY*NLEVELS; j++)  /* new */
		if (!(j % NLEVELS))
		{
			for (k = 0; k < NLEVELS; k++)
				ficache[j+k].c_level = fdcache[j+k].c_level = k + 3;
		}

	init_my_statistic_arrays();

	/*init Tlb size */

	/*init_Tlb();  extra */
	if (HOW_MANY)
	{
		for (i = 0; i < HOW_MANY + 1; i++)
		{
			InstrDTBSIZE[i] = InstrTLBParams[i];
			DTBSIZE[i] = DataTLBParams[i];
		}
	}

	for (i = 0; i < 3; i++)
	{
		int	sim_count = 0;
		char	tmp_s[100];

		if (i == 0)
		{
			a = 'i';

			if (!HOW_MANY)
			{
				/*
				fprintf(stderr,"InstrCacheSize = %d \n",InstrCacheSize);
				sprintf(tmp_s,"-i%dKb64,32s5rlruwt",InstrCacheSize);
				fprintf(stderr,"InstrCacheParams[0] is %d \n",InstrCacheParams[0]);
				*/
				sprintf(tmp_s,"-i%dKb64,32s5rlruwt",InstrCacheParams[0]);
				/*
				fprintf(stderr,"1.1 tmp_s is %s \n",tmp_s);
					strcpy(tmp_s,"-i5Kb64,16s5rlruwt");
				*/
			}
			else
			{
/*
				fprintf(stderr,"InstrCacheParams[0] is %d \n",InstrCacheParams[0]);
*/

				sprintf(tmp_s,"-i%d%s64,32s5rlruwt",InstrCacheParams[0],InstrCacheWhat[0]);
/*
				fprintf(stderr,"1.2 tmp_s is %s \n",tmp_s);
*/
			}
		}
		else	
		if (i == 1)
		{
			a = 'd';
			
			if (!HOW_MANY)
			{
				sprintf(tmp_s,"-d%d%s32s4rlruwtwa",DataCacheParams[0],DataCacheWhat[0]);
				/*
				sprintf(tmp_s,"-d%dKb32s4rlruwtwa",DataCacheSize);
					strcpy(tmp_s,"-d256Kb16s4rlru");
				*/
			}
			else
			{
				sprintf(tmp_s,"-d%d%s32s4rlruwtwa",DataCacheParams[0],DataCacheWhat[0]);
/*
				fprintf(stderr,"2 tmp_s is %s \n",tmp_s);
*/
			}
		}
		else
		if (i == 2)
		{
			a = 'u';
		
			if (!HOW_MANY)
			{
				sprintf(tmp_s,"-u%d%s128,32wbIi1Id1",SecLevelCacheParams[0],SecLevelCacheWhat[0]);
				/*
				sprintf(tmp_s,"-u%dKb128,32wbIi2Id2",SecondLeveCacheSize);
					strcpy(tmp_s,"-u4Mb128,32wbwaIi1Id1");
				*/
			}
			else
			{
				sprintf(tmp_s,"-u%d%s128,32wbIi1Id1",SecLevelCacheParams[0],SecLevelCacheWhat[0]);
/*
				fprintf(stderr,"3 tmp_s is %s \n",tmp_s);
*/
			}
		}

/*
        fprintf(stderr,"cachesim5.c: 320\n");
*/
		switch (a) 
		{

			case 'd':
			case 'i':
				if (id && id != a)
					shade_fatal ("unmatched -i/-d");
				if (nlevels >= NLEVELS)
					shade_fatal ("too many cache levels");
				c = (a == 'i' ? icaches : dcaches) + nlevels;
				newcache (c,tmp_s,sim_count++,0); 
				for (j = 0; j < HOW_MANY*NLEVELS; j++)  /* new */
					if (!(j % NLEVELS))
					{
						c = (a == 'i' ? ficache+j : fdcache+j) + nlevels;
						if (i == 0)
						{
							sprintf(tmp_s,"-i%d%s64,32s5rlruwt",InstrCacheParams[j/NLEVELS+1],InstrCacheWhat[j/NLEVELS+1]);
/*
							fprintf(stderr,"System %d : tmp_s = %s \n",j/NLEVELS+1,tmp_s);
*/
						}
						else
						if (i == 1) /* or a == 'd' */
						{
							sprintf(tmp_s,"-d%d%s32s4rlruwtwa",DataCacheParams[j/NLEVELS+1],DataCacheWhat[j/NLEVELS+1]);
/*
							fprintf(stderr,"System %d : tmp_s is %s \n",j/NLEVELS+1,tmp_s);
*/
						}
						newcache (c, tmp_s,sim_count++,j);
					}

				if (id == a) 
				{
					nlevels++;
					id = 0;
				}
				else	
					id = 'i' + 'd' - a;
				break;
			case 'u':
				if (id)
					shade_fatal ("unmatched -i/-d");
				if (nlevels >= NLEVELS)
					shade_fatal ("too many cache levels");
				newcache (c = icaches + nlevels,tmp_s,sim_count++,0);
				dcaches[nlevels] = *c; /* to nlevels++ tha ginei parakatw */

				for (j = 0; j < HOW_MANY*NLEVELS; j++)  /* new */
					if (!(j % NLEVELS))
					{
						sprintf(tmp_s,"-u%d%s128,32wbwaIi1Id1",SecLevelCacheParams[j/NLEVELS+1],SecLevelCacheWhat[j/NLEVELS+1]);
/*
						fprintf(stderr,"System %d : tmp_s is %s \n",j/NLEVELS+1,tmp_s);
*/
						newcache (c = ficache+j + nlevels, tmp_s,sim_count++,j);
						fdcache[j+nlevels] = *c;
					}
				nlevels++; /* new ... here ... */

				break;
			default:
/*
				fprintf(stderr,"Pali malakia ekana ... \n");
*/
				usage ("");
				break;
		}
	}

	if (id)
		shade_fatal ("unmatched -i/-d");
	if (nlevels <= 0)
		usage ("no caches");

	for (i = 0; i < nlevels-1; i++) 
	{
		icaches[i].c_next = icaches + i + 1;
		dcaches[i].c_next = dcaches + i + 1;
	}
	icaches[nlevels-1].c_next = dcaches[nlevels-1].c_next = 0;

	for (j = 0; j < HOW_MANY*NLEVELS; j++) /* new */
	{
		int     tmp;

		if (!(j % NLEVELS))
		{
			for (k = 0; k < nlevels-1; k++)
			{
				tmp = j+k;
				ficache[tmp].c_next = ficache + j+ k + 1;
				fdcache[tmp].c_next = fdcache + j+ k + 1;
			}
			ficache[j+ nlevels-1].c_next = fdcache[j+ nlevels-1].c_next = 0;
		}
	}

	init_memsize();

	if (!first_time_only)
	{
		abel_file = fopen("/tmp/ABEL_ddtb","w");
		first_time_only = 1;

		(void) shade_trctl_trsize (sizeof (Trace));
		(void) shade_trctl_it ((uint32) IT_ANY, 1, 1, 
	    	(uint32) (TC_ANNULLED | TC_IH | TC_PC));
		(void) shade_trctl_it ((uint32) (IT_LOAD | IT_STORE), 1, 1, 
	    	(uint32) (TC_ANNULLED | TC_IH | TC_PC | TC_EA));
	}

/**************
	ReadLineFile("../interface/SC/ADDRESS");
**************/
}

struct timeval  tp, tp_final;
struct timezone tzp, tzp_final;

/* danalis code START */

/*
extern void Change_Contents();
extern void MarkLine();
*/

extern char *shade_bench_path;
int IFoundMain = 0;

/* danalis code END */

int analyze(flag)
int	flag;
{
  register Trace *tr;
  register int    ih;
  int             j, k;

/**** danalis code START ****/
  FILE *pipe_in, *pipe_out;
  FILE *pipe4gdb;
  char *str4gdb;          /* The command that will give us the pc of main */
  char gdbprompt[64];
  int  c, danI = 0, main_pc = -1, GdbOk = 1;
  int  readfd[2], writefd[2];
  int  interfpid;

/*
  fprintf(stderr,"analyze() START\n");
  printf("analyze() START\n");
*/
  fflush(stdout);
  /**** Get our parents (interf) pid so we can send ****/
  /**** him a signal when we do all we should       ****/
  interfpid = getppid();

  /**** Make the string to feed popen() ****/
  str4gdb = (char *)calloc(64+strlen(shade_bench_path), sizeof(char));
  strcat(str4gdb,"echo 'printf \"0x%x\\n\",main' | gdb -q ");
  strcat(str4gdb,shade_bench_path);
  strcat(str4gdb," 2> /dev/null");
  
  /**** Print some messages ******
  fprintf(stderr,"\nDanMsg: Application: #%s#\n",shade_bench_path);
  fprintf(stderr,"\nDanMsg: Running: %s\n",str4gdb);
  *******************************/

  /**** Run gdb and read the answer ****/
  pipe4gdb = popen(str4gdb, "r");

  if(pipe4gdb == NULL)
    fprintf(stderr,"popen returned a NULL pointer !!!\n");
  if(pipe4gdb < 0)
    fprintf(stderr,"popen returned -1, command may be malformed\n");
  /*********
  if(pipe4gdb)
    fprintf(stderr,"popen returned normally\n");

  fprintf(stderr,"Scaning pipe for popen's answer ...\n");
  *********/
  fscanf(pipe4gdb, "%s 0x%x",gdbprompt, &main_pc);          
  /***** a hack due to gcc strange behavior *****/
  while( ( c=getc(pipe4gdb) ) != EOF);
  /******
  fprintf(stderr,"main is at: 0x%x\n",main_pc);  
  ******/
  pclose(pipe4gdb);

  /**** check if the answer is what we expected ****/
  if( strcmp(gdbprompt,"(gdb)") ){
    fprintf(stderr,"gdb cannot locate main, probably %s",shade_bench_path);
    fprintf(stderr," was not compiled with -g option\n");
    GdbOk = 0;
  }
  free(gdbprompt); 
  free(str4gdb); 

  /**** make the pipes that we 'll use to communicate with addr2line ****/
  if( pipe(readfd) ){   /**** the kid shall read from readfd[0]      ****/
    perror("pipe read");
  }
  if( pipe(writefd) ){  /**** and write to writefd[1]                ****/
    perror("pipe write");
  }

  if( fork() ){  /**** The father has all the original code of analyze ****/
                 /**** plus the routine that does the communication    ****/

    char a2lAnswer[512];  /**** To store the answer of addr2line       ****/
    char *WhichFile;      /**** To store the filename                  ****/
    char *fullpath;
    char *OldFile;
    char *LineNumber;     /**** To store the Line Number               ****/
    char tmp[PATH_MAX+256], *lastslash;
    char *str4fifo;       /**** To write to the named pipe             ****/
    int posInBuf = 0;
    int fifofd;           /**** File descriptor for the named pipe     ****/
    struct timeval tp4w;
    struct timezone tpz4w;
   
    /*WhichFile = (char *)calloc(512*sizeof(char)); */
    OldFile   = (char *)calloc(512*sizeof(char));
    str4fifo  = (char *)calloc(512*sizeof(char));

    /**** Close the edges that kid will use ****/
    if( close(readfd[0]) < 0 )
      perror("Father: close read");
    if( close(writefd[1]) < 0 )
      perror("Father: close write");

    /**** Get file pointers from the fildes, to work with getc() ****/
    if( (pipe_out = fdopen(readfd[1],"w"))  == NULL )
      perror("Father: fdopen read");
    if( (pipe_in  = fdopen(writefd[0],"r")) == NULL )
      perror("Father: fdopen write");

    /**** Create a named pipe to communicate with Interf ****/
    unlink("/tmp/.fifo");
    if( mkfifo("/tmp/.fifo", S_IWUSR|S_IRUSR) ){
      perror("mkfifo");
    }

    /**** Open the named pipe ****/
    if( (fifofd = open("/tmp/.fifo", O_WRONLY)) == -1 ){
      perror("open fifo");
    }

/**** danalis code END ****/

    resetcaches();

    REPEATS = 1000;
	
/**
    if (flag == 0)
      fprintf(stderr," Analyze() is called with flag 0\n");
    fprintf(stderr,"analyze: HOW_MANY = %d \n",HOW_MANY);
**/
    if (flag)
      while (tr = shade_step()){

/****** danalis code START ******/
        danI++;
        if( GdbOk && tr->tr_pc == main_pc ){
/***
          fprintf(stderr,"main() started\n");
***/
          IFoundMain = 1;
        }

        /* feed addr2line with an address */
        if( !(danI%2503) && IFoundMain ){
          fprintf(pipe_out,"%x\n",tr->tr_pc);
          fflush(pipe_out);
     
          /* Read the answer and store it to tmp[] */
          posInBuf = 0;
          while( (c = getc(pipe_in)) != EOF ){
           tmp[posInBuf++] = c;
/****           putc(c,stderr); ****/
           if( c == '\n' )
             break; 
          }
          tmp[posInBuf] = '\0';
          lastslash = strrchr(tmp,'/')+1;
/*          strcpy(WhichFile,lastslash); */
          LineNumber = strrchr(tmp,':');
          *LineNumber = '\0';
          WhichFile = lastslash;
          fullpath  = tmp;

/**** Make the string we'll write to the fifo ****/
          strcpy(str4fifo, fullpath);
          strcat(str4fifo," ");
          strcat(str4fifo, LineNumber+1);

/**** If really a source file ****/
          if( !strstr(WhichFile,"libg.s") ){
/********
            fprintf(stderr,"%s",WhichFile);
            fprintf(stderr," %d\n",atoi(LineNumber+1));
            fprintf(stderr,"%s\n",str4fifo);
********/
            if( write( fifofd, str4fifo, strlen(str4fifo) ) <= 0 )
              perror("in fork()'s father, in analyze()");
          }

          posInBuf = 0;
          danI = 0;
        }

/****** danalis code END ******/

        accesscache (icaches, tr->tr_pc, R, 2);

        if (tr->tr_annulled)
          goto THERE;
        else
          nina++;
  
        ih = tr->tr_ih;

        if (ih_isload (ih)) {
          accesscache (dcaches, tr->tr_ea, R, ihash_memsize[ih]);
        }
        else{
          if (ih_isstore (ih))
            accesscache (dcaches, tr->tr_ea, W, ihash_memsize[ih]);
        }

		
THERE:
        if (HOW_MANY){
          j = 0;
          while (j < HOW_MANY*NLEVELS){
            accesscache (ficache[j], tr->tr_pc, R, 2);
            if (tr->tr_annulled)
              goto MY_CONT_1;
            else
              nina++;

            ih = tr->tr_ih;
            if (ih_isload (ih))
              accesscache (fdcache[j], tr->tr_ea, R, ihash_memsize[ih]);
            else
              if (ih_isstore (ih))
                accesscache (fdcache[j], tr->tr_ea, W, ihash_memsize[ih]);

MY_CONT_1:
           j+=NLEVELS;
          }
        }
      }
    close(fifofd);
    unlink("/tmp/.fifo");
	
	/*************** milolid ************************************/
	/******** Send mesg that there are not any more data *******/
	
      if (CONNECTION_IS_OK && (!HOW_MANY) && !READ_COMMANDS_FROM_FILE){
        PrintUsefulResults();
        if (abel_file)
          fclose(abel_file);
        Update_Interface_Info.kind_of_mesg = -1; /* means stop sending data */
        CONNECTION_IS_OK = 0;
        write(connection,&Update_Interface_Info,sizeof(Msg_Info));
	  }
	  else
        if (CONNECTION_IS_OK && HOW_MANY && !READ_COMMANDS_FROM_FILE){
          FILE  *myprintFp;
  
          PrintUsefulResults();
          if (abel_file)
            fclose(abel_file);
          myprintFp = fopen("/tmp/TOTAL_STATS","w");
          
          if (myprintFp){
            cachestats(myprintFp, ficache, fdcache, nina);
            fclose(myprintFp);
          }
          CONNECTION_IS_OK = 0; /* k.l.p. */
          Update_Simultaneous_Interface.kind_of_mesg = -1;
          write(connection,&Update_Simultaneous_Interface,sizeof(Msg_simultaneous));
        }
	/***************************************** milolid **********/

#if 0	/* kelly says don't bother; hardware may not do this ... */
  cleancaches();
  gettimeofday(&tp_final,&tzp_final);
  fprintf(stderr,"RUNS FOR %d sec \n",tp_final.tv_sec - tp.tv_sec);
#endif

/**** danalis code START  ****/
  }
  else{ /* The kid just links stdout and stdin   */
        /* to the pipes and then calls addr2line */
    char str4system[PATH_MAX+256];

    strcpy(str4system, "addr2line -e ");
    strcat(str4system, shade_bench_path);

    /**********************
    fprintf(stderr,"*** system(%s)\n",str4system);
    fprintf(stderr,"************************\n");
    fprintf(stderr,"sending SIGUSR1 to: %d\n",interfpid);
    fprintf(stderr,"************************\n");
    **********************/
    if( kill(interfpid, SIGUSR1) < 0)
      perror("********* kill");

    if(close(readfd[1]) < 0)
      perror("Child: close read");
    if(close(writefd[0]) < 0)
      perror("Child: close write");
    if( dup2(readfd[0],0) < 0 )
      perror("Child: dup2 read");
    if( dup2(writefd[1],1) < 0 )
      perror("Child: dup2 write");
    system(str4system); 

  }
/**** danalis code END  ****/

  return (0);
}

FILE	*myprintFp;

void
terminate()
{
	int	i;
	
	if (abel_file)
		fclose(abel_file);
	
	myprintFp = fopen("/tmp/ABEL_TOTAL_STATS","w");
	if (myprintFp)
	{
/***
		fprintf(stderr,"nina is %d \n",nina);
***/
		cachestats (myprintFp, icaches, dcaches, nina);
		cachestats (myprintFp, ficache, fdcache, nina);

		fclose(myprintFp);
	}
	DtbPrintResults(); /* milolid : Prints Dtb results in a .out file */

	/*************************** milolid ************************/
	fprintf(stderr,"L1_I_Cache_miss = %d ,\n L1_D_Cache_miss = %d ,\n L2_Cache_miss = %d \n",L1_I_Cache_miss, L1_D_Cache_miss, L2_Cache_miss);
	fprintf(stderr,"L1_I_Total_References = %d ,\n L1_D_Total_References = %d ,\n L2_Total_References = %d \n", L1_I_Total_References, L1_D_Total_References, L2_Total_References);

	for (i = 0; i < HOW_MANY; i++)
	{
		fprintf(stderr,"array_L1_I_Cache_miss[%d] = %d \n",i+1,array_L1_I_Cache_miss[i]);
		fprintf(stderr,"array_L1_D_Cache_miss[%d] = %d \n",i+1,array_L1_D_Cache_miss[i]);
		fprintf(stderr,"array_L2_Cache_miss[%d] = %d \n",i+1,array_L2_Cache_miss[i]);

		fprintf(stderr,"array_L1_I_Total_References[%d] = %d \n",i+1,array_L1_I_Total_References[i]);
		fprintf(stderr,"array_L1_D_Total_References[%d] = %d \n",i+1,array_L1_D_Total_References[i]);
		fprintf(stderr,"array_L2_Total_References[%d] = %d \n",i+1,array_L2_Total_References[i]);
	}
	/*************************** milolid ************************/
	fprintf(stderr,"Cachesim5: RUNS FOR %d sec \n",tp_final.tv_sec - tp.tv_sec);
}


/*
 * access cache c for reading (rw=R) or writing (rw=W)
 * 1<<size bytes at address a.  a is presumed to be 
 * aligned on 1<<size boundary (size lsb's of a == 0).
 */

static void
accesscache (c, a, rw, size)
Cache	*c;
addr_t	a;
int	rw, size;
{
	int	bsize, bbsize, ssize, s, rwalloc;
	int	line_no = 1;
	int	which_window = 1;
	addr_t	T, X;
	addr_list	*addr_point;
	mask_t	Y;
	static int flag = 0,flag2 = 0;

top:
	/***** H Dtb routines, den epistrefoun fusikh dieuthunsh
	       kathws oi to mapping pou apofasisame einai, h ideath
	       dieuthunsh X na antistoixei sthn X fusikh dieuthunsh.
	******/

	addr_point = search_addr_list(a);
	if (!addr_point)
	{
		/*
		fprintf(stderr,"I cannot find address in addr list \n");
		line_no = 0;
		which_window = -1;
		*/
		;
	}
	else
	{
		line_no = addr_point->line;
		which_window = addr_point->which_window;
		/*
		fprintf(stderr,"Line is %d (window %d)\n",line_no, which_window);
		*/
	}

	if (c->c_idu == 'I')
	{
		InstrDtbReference(a,c->c_sim);
	}
	else
	if (c->c_idu == 'D')
	{
		DtbReference(a,c->c_sim);
	}
	else
	{
		if (!flag2)
		{
			flag2 = 1;
/*
			fprintf(stderr,"nothing ....c_idu = %c \n",c->c_idu);
*/
		}
	}

	/*
	printf("before milolid_main ... a = 0x%lx \n",a);
	milolid_main(a);
	*/

	if (c->c_idu == 'I' || c->c_idu == 'D')
	{
		if (c->c_idu == 'I')
		{
			if (c->c_sim == 0)
				L1_I_Total_References++;
			else
				array_L1_I_Total_References[c->c_sim - 1]++;
		}
		else /* mipws prepei na elegxw kai gia D ti c_idu ??? what else it can be ? */
		{
			if (c->c_sim == 0)
				L1_D_Total_References++;
			else
				array_L1_D_Total_References[c->c_sim - 1]++;
		}
	}
	else
	if (c->c_idu == 'U')
	{
		if (c->c_sim == 0)
			L2_Total_References++;
		else
			array_L2_Total_References[c->c_sim - 1]++;
	}
	else
        ;
		/*
		fprintf(stderr,"What is this really\n");
		fprintf(stderr,"Unexpected Level %d.\n",c->c_level);
		*/
	if (c->c_idu == 'I' || c->c_idu == 'D') /*new ... */
		TOTAL_COUNTER[c->c_sim]++;
	if (TOTAL_COUNTER[c->c_sim] == REPEATS && CONNECTION_IS_OK)
	{
		TOTAL_COUNTER[c->c_sim] = 0;

		if (0/*CONNECTION_IS_OK && (!HOW_MANY)*/)
		{
			fprintf(abel_file,"%d\n",array_misses[0]);
			Update_Interface_Info.ddtb_misses = array_misses[0] - ex_misses;
			Update_Interface_Info.ddt_ref = array_hits[0]+array_misses[0] - ex_misses - ex_hits;
			Update_Interface_Info.idtb_misses = array_InstrMisses[0] - ex_InstrMisses;
			Update_Interface_Info.idtb_ref = array_InstrHits[0] + array_InstrMisses[0] - ex_InstrMisses - ex_InstrHits;
			Update_Interface_Info.d_cache_misses = L1_D_Cache_miss - ex_L1_D_Cache_miss;
			Update_Interface_Info.d_cache_ref = L1_D_Total_References - ex_L1_D_Total_References;
			Update_Interface_Info.i_cache_misses = L1_I_Cache_miss - ex_L1_I_Cache_miss;
			Update_Interface_Info.i_cache_ref = L1_I_Total_References - ex_L1_I_Total_References;
			Update_Interface_Info.u_cache_misses = L2_Cache_miss - ex_L2_Cache_miss;
			Update_Interface_Info.u_cache_ref = L2_Total_References - ex_L2_Total_References;
			
			Update_Interface_Info.kind_of_mesg = 0; /* if -1, means no more data */
			Update_Interface_Info.line_no = line_no;
			Update_Interface_Info.which_window = which_window;

			ex_misses = array_misses[0];
			ex_hits = array_hits[0];
			ex_InstrMisses = array_InstrMisses[0];
			ex_InstrHits = array_InstrHits[0];
			ex_L1_D_Cache_miss = L1_D_Cache_miss;
			ex_L1_D_Total_References = L1_D_Total_References;
			ex_L1_I_Cache_miss = L1_I_Cache_miss;
			ex_L1_I_Total_References = L1_I_Total_References;
			ex_L2_Cache_miss = L2_Cache_miss;
			ex_L2_Total_References = L2_Total_References;

			write(connection,&Update_Interface_Info,sizeof(Msg_Info));
		}
		else
		if (CONNECTION_IS_OK /*&& HOW_MANY*/)	/* multiple simultaneous executions */
		{
			Update_Simultaneous_Interface.ddtb_misses = array_misses[c->c_sim] - array_ex_misses[c->c_sim];
			Update_Simultaneous_Interface.ddt_ref = array_misses[c->c_sim] + array_hits[c->c_sim] - array_ex_misses[c->c_sim] - array_ex_hits[c->c_sim];

			Update_Simultaneous_Interface.idtb_misses = array_InstrMisses[c->c_sim] - array_ex_InstrMisses[c->c_sim];
			Update_Simultaneous_Interface.idtb_ref = array_InstrMisses[c->c_sim] + array_InstrHits[c->c_sim] - array_ex_InstrMisses[c->c_sim] - array_ex_InstrHits[c->c_sim];
			if (c->c_sim == 0)
			{
				Update_Simultaneous_Interface.d_cache_misses = L1_D_Cache_miss - ex_L1_D_Cache_miss;
				Update_Simultaneous_Interface.d_cache_ref = L1_D_Total_References - ex_L1_D_Total_References;
			}
			else
			{
				Update_Simultaneous_Interface.d_cache_misses = array_L1_D_Cache_miss[c->c_sim-1] - array_ex_L1_D_Cache_miss[c->c_sim-1];
				Update_Simultaneous_Interface.d_cache_ref = array_L1_D_Total_References[c->c_sim-1] - array_ex_L1_D_Total_References[c->c_sim-1];
			}

			if (c->c_sim == 0)
			{
				Update_Simultaneous_Interface.i_cache_misses = L1_I_Cache_miss - ex_L1_I_Cache_miss;
				Update_Simultaneous_Interface.i_cache_ref = L1_I_Total_References - ex_L1_I_Total_References;
			}
			else
			{
				Update_Simultaneous_Interface.i_cache_misses = array_L1_I_Cache_miss[c->c_sim-1] - array_ex_L1_I_Cache_miss[c->c_sim-1];
				Update_Simultaneous_Interface.i_cache_ref = array_L1_I_Total_References[c->c_sim-1] - array_ex_L1_I_Total_References[c->c_sim-1];
			}

			if (c->c_sim == 0)
			{
				Update_Simultaneous_Interface.u_cache_misses = L2_Cache_miss - ex_L2_Cache_miss;
				Update_Simultaneous_Interface.u_cache_ref = L2_Total_References - ex_L2_Total_References;
				/*Update_Simultaneous_Interface.u_cache_ref = L2_Total_References;*/
			}
			else
			{
				Update_Simultaneous_Interface.u_cache_misses = array_L2_Cache_miss[c->c_sim-1] - array_ex_L2_Cache_miss[c->c_sim-1];
				Update_Simultaneous_Interface.u_cache_ref = array_L2_Total_References[c->c_sim-1] - array_ex_L2_Total_References[c->c_sim-1];
			}

			Update_Simultaneous_Interface.c_sim = c->c_sim;
			Update_Simultaneous_Interface.kind_of_mesg = 0;
			Update_Simultaneous_Interface.line_no = line_no;
			/*
			fprintf(stderr,"I send which_window to be %d \n",which_window);
			*/
			Update_Simultaneous_Interface.which_window = which_window;

			array_ex_misses[c->c_sim] = array_misses[c->c_sim];
			array_ex_InstrMisses[c->c_sim] = array_InstrMisses[c->c_sim];
			array_ex_hits[c->c_sim] = array_hits[c->c_sim];
			array_ex_InstrHits[c->c_sim] = array_InstrHits[c->c_sim];
			if (c->c_sim == 0)
			{
				ex_L1_D_Cache_miss = L1_D_Cache_miss;
				ex_L1_I_Cache_miss = L1_I_Cache_miss;
				ex_L2_Cache_miss = L2_Cache_miss;
				ex_L1_D_Total_References = L1_D_Total_References;
				ex_L1_I_Total_References = L1_I_Total_References;
				ex_L2_Total_References = L2_Total_References;
			}
			else
			{
				array_ex_L1_I_Cache_miss[c->c_sim-1] = array_L1_I_Cache_miss[c->c_sim-1];
				array_ex_L1_D_Cache_miss[c->c_sim-1] = array_L1_D_Cache_miss[c->c_sim-1];
				array_ex_L2_Cache_miss[c->c_sim-1] = array_L2_Cache_miss[c->c_sim-1];
				array_ex_L1_D_Total_References[c->c_sim-1] = array_L1_D_Total_References[c->c_sim-1];
				array_ex_L1_I_Total_References[c->c_sim-1] = array_L1_I_Total_References[c->c_sim-1];
				array_ex_L2_Total_References[c->c_sim-1] = array_L2_Total_References[c->c_sim-1];
			}

			write(connection,&Update_Simultaneous_Interface,sizeof(Msg_simultaneous));
		}
		else
          ;
/*
			fprintf(stderr,"Ti malakas pou eimai !!!!\n");
*/
	}

/***************** milolid *********************/

	ssize = c->c_ssize;	/* set size */
	bsize = c->c_bsize;	/* block size */
	bbsize = c->c_bbsize;	/* subblock size */

	rwalloc = rw == R || c->c_walloc;

	if (size > bbsize) 
	{
		addr_t	ainc, alim = a + (1 << size);

		if (bbsize < 0) 
		{
			/*
			 * null cache (hole in cache system).
			 */
			c->c_nref[size][rw]++;
			if (c->c_next)
				accesscache (c->c_next, a, rw, size);
			return;
		}

		for (ainc = 1 << bbsize; a < alim; a += ainc)
			accesscache (c, a, rw, bbsize);
		return;
	}

	if (0)
	printf("Before add nref c_sim is %d and nref are %lf with size %d and rw %d\n",c->c_sim,c->c_nref[size][rw],size,rw);
	c->c_nref[size][rw]++;
	if (0)
	printf("1 After add nref c_sim is %d and nref are %lf with size %d and rw %d\n",c->c_sim,c->c_nref[size][rw],size,rw);

	T = a >> ssize;				/* tag */
	X = (a & sizemask (ssize)) >> bsize;	/* block # (index) */
	Y = (a & sizemask (bsize)) >> bbsize;	/* subblock # */
	Y = 1 << Y;				/* subblock bit mask */

	for (s = 0; s < c->c_nsets; s++)
		if (c->c_tags[s][X] == T) 
		{
			if (c->c_valid[s][X] & Y) 
			{
				if (c->c_nsets > 1)
					accesscacheset (c, s, X);
				if (rw == R)
					return;
				else	goto hit;
			}
			else 
			{
				/*
				 * nonsubblocked caches don't distinguish tag
				 * mismatch misses from unset valid bit misses.
				 */
				if (bsize == bbsize)
					continue;
				goto miss;
			}
		}

	if (rwalloc) 
	{
		if (c->c_nsets > 1)
			s = replacecacheset (c, X);
		else	s = 0;

		if (c->c_valid[s][X]) 
		{
			if (!c->c_wthru)
				cleancacheline (c, s, X);

			if (c->c_bkinv)
				backinvalidatingcache (c, s, X);

			c->c_valid[s][X] = 0;
		}
		c->c_tags[s][X] = T;
	}
	else	/*
		 * on write miss to virtual, write thru, no write alloc cache,
		 * invalidate subblock when tag mismatches in case of aliasing.
		 * only implemented for 1st level, direct mapped caches.
		 */
		if (c->c_winv)
			if (c->c_valid[0][X] & Y) 
			{
				c->c_valid[0][X] &= ~Y;    /* c_bkinv==0 */
				c->c_nwinv++;
			}

miss:
	/******* milolid - misses ...***************/
	if (c->c_idu == 'I')
	{
		if (c->c_sim == 0)
			L1_I_Cache_miss++;
		else
			array_L1_I_Cache_miss[c->c_sim - 1]++;
	}
	else
	if (c->c_idu == 'D')
	{
		if (c->c_sim == 0)
			L1_D_Cache_miss++;
		else
			array_L1_D_Cache_miss[c->c_sim - 1]++;
	}
	else
	if (c->c_idu == 'U')
	{
		if (c->c_sim == 0)
			L2_Cache_miss++;
		else
			array_L2_Cache_miss[c->c_sim - 1]++;
	}

	/********* milolid ************/

	c->c_nmiss[size][rw]++;

	if (c->c_next && rw == W && c->c_walloc)
		accesscache (c->c_next, align (a, bbsize), R, bbsize);

	if (rwalloc) 
	{
		c->c_valid[s][X] |= Y;
		c->c_used[s][X] = 1;
		if (c->c_nsets > 1)
			accesscacheset (c, s, X);
	}

	if (rw == W && !c->c_wthru)
		c->c_dirty[s][X] |= Y;
	else	
		if (c->c_next)
			goto nextlevel;

	return;

hit:
	if (c->c_wthru) 
	{
		if (c->c_next)
			goto nextlevel;
	}
	else 
	{
		if (c->c_dirty[s][X] == 0)
			c->c_nrw++;
		c->c_dirty[s][X] |= Y;
	}

	return;

nextlevel:
	/*
	 * accesscache (c->c_next, align (a, bbsize), rw, bbsize);
	 * (tail recursion elimination)
	 */
	c = c->c_next;
	a = align (a, bbsize);	/* in case bbsize > c->c_next->c_bbsize */
	size = bbsize;
	goto top;
}


static void
accesscacheset (c, s, X)
Cache	*c;
int	s;
addr_t	X;
{
	switch (c->c_replace) 
	{
		case C_LRU:
			if (c->c_lru[s][X] = ++c->c_lru[-1][X])
				break;
			/*
		 	* minor glitch on lru counter overflow ...
		 	*/
			for (s = -1; s < c->c_nsets; s++)
				c->c_lru[s][X] = 0;
			break;
		case C_RANDOM:
			break;
		case C_VIKING:
			c->c_mru[X] |= 1 << s;
			if (c->c_mru[X] == sizemask (c->c_nsets))
				c->c_mru[X] = 1 << s;
			break;
	}
}


/*
 * invalidate any data in cache c in address range [a,a+1<<size)
 */

static void
backinvalidatedcache (c, a, size)
Cache	*c;
addr_t	a;
int	size;
{
	int	bsize, bbsize, ssize, s;
	addr_t	a2, ainc, ainc2, alim, T, X;
	mask_t	Y, v;

	ssize = c->c_ssize;	/* set size */
	bsize = c->c_bsize;	/* block size */
	bbsize = c->c_bbsize;	/* subblock size */

	T = a >> ssize;				/* tag */
	X = (a & sizemask (ssize)) >> bsize;	/* block # (index) */

	if (size < bsize)
		goto small;

	/*
	 * invalidate a block at a time.
	 */

	alim = a + (1 << size);
	ainc = 1 << bsize;
	ainc2 = 1 << bbsize;

	for (; a < alim; a += ainc, X++)
		for (s = 0; s < c->c_nsets; s++) 
		{
			if (T != c->c_tags[s][X])
				continue;

			v = c->c_valid[s][X];

			for (a2 = a; a2 < a+ainc; a2 += ainc2, v >>= 1)
				if (v & 1) 
				{
					c->c_nbkinved++;
					if (c->c_bkinv)
						backinvalidatingcachesubblock
						    (c, a2);
				}

			c->c_valid[s][X] = 0;
		}
	return;
	
small:
	/*
	 * invalidate a subblock at a time.
	 * inefficient below, but unlikely case.
	 */

	if (size < bbsize) 
	{
		size = bbsize;			/* invalidate whole subblock */
		a = (a >> size) << size;	/* realign */
	}

	Y = (a & sizemask (bsize)) >> bbsize;	/* subblock # */
	Y = 1 << Y;				/* subblock bit mask */

	alim = a + (1 << size);
	ainc = 1 << bbsize;

	for (; a < alim; a += ainc, Y <<= 1)
		for (s = 0; s < c->c_nsets; s++)
			if (T == c->c_tags[s][X])
				if (c->c_valid[s][X] & Y) 
				{
					c->c_nbkinved++;
					c->c_valid[s][X] &= ~Y;
					if (c->c_bkinv)
						backinvalidatingcachesubblock
						    (c, a);
				}
	return;
}


/*
 * for each valid subblock in set s, line X of cache c
 * generate a back invalidate request to each cache
 * included by cache c.
 */

static void
backinvalidatingcache (c, s, X)
Cache	*c;
int	s;
addr_t	X;
{
	addr_t	a, ainc, alim;
	mask_t	v;

	a = (c->c_tags[s][X] << c->c_ssize) | (X << c->c_bsize);
	ainc = 1 << c->c_bbsize;
	alim = a + (1 << c->c_bsize);

	v = c->c_valid[s][X];

	for (; a < alim; a += ainc, v >>= 1)
		if (v & 1)
			backinvalidatingcachesubblock (c, a);

}


static void
backinvalidatingcachesubblock (c, a)
Cache	*c;
addr_t	a;
{
	Cache	**bc;

	c->c_nbkinving++;

	for (bc = c->c_bkinv; *bc; bc++)
		backinvalidatedcache (*bc, a, c->c_bbsize);
}


#if 0
/*
 * write back dirty subblocks in cache c.
 */

static void
cleancache (c)
Cache	*c;
{
	int	s;
	addr_t	X;

	if (!c->c_wthru && c->c_walloc)
		for (s = 0; s < c->c_nsets; s++)
			for (X = 0; X < c->c_nlines; X++)
				cleancacheline (c, s, X);
}
#endif


/*
 * write back dirty subblocks in set s, line X of cache c.
 */

static void
cleancacheline (c, s, X)
Cache	*c;
int	s;
addr_t	X;
{
	int	bsize, bbsize;
	addr_t	a, ainc, alim;
	mask_t	vd, Y;

	vd = c->c_valid[s][X] & c->c_dirty[s][X];
	c->c_dirty[s][X] = 0;

	if (vd == 0)
		return;

	bsize = c->c_bsize;
	bbsize = c->c_bbsize;

	a = (c->c_tags[s][X] << c->c_ssize) | (X << bsize);
	alim = a + (1 << bsize);
	ainc = 1 << bbsize;

	for (Y = 1; a < alim; a += ainc, Y <<= 1)
		if (vd & Y) 
		{
			c->c_nwback++;
			if (c->c_next)
				accesscache (c->c_next, a, W, bbsize);
		}
}


#if 0
static void
cleancaches()
{
	int	i;

	for (i = 0; i < nlevels; i++) 
	{
		/*
		 * do d before i so that for unified write back caches,
		 * the write backs etc are charged against the d cache.
		 */
		cleancache (dcaches + i);
		cleancache (icaches + i);
	}
}
#endif


static void
invalidatecache (c)
Cache	*c;
{
	int	s;
	addr_t	X;

	if (c->c_valid)
		for (s = 0; s < c->c_nsets; s++)
			for (X = 0; X < c->c_nlines; X++)
				c->c_valid[s][X] = 0;
 
	 if (c->c_dirty)
		 for (s = 0; s < c->c_nsets; s++)
			 for (X = 0; X < c->c_nlines; X++)
				 c->c_dirty[s][X] = 0;
}


/*
 * -{i|d|u}	select instruction, data, unified cache
 * <sz>		overall cache size in bytes (suffix K for 1024x, etc.)
 * b<bsz>[,<bbsz>]  block and subblock sizes in bytes
 * [s<#sets>]	set associativity (1 for direct mapped (default))
 * [r<replace>]	set replacement: lru, random, or viking
 * [wt|wb]	write thru or write back
 * [wa]		write allocate
 * [wi]		write invalidate
 * [I<cache>]	include cache (e.g. i1 for 1st level I cache)
 */

static void
newcache (c, s,c_sim,offset)
Cache	*c;
char	*s;
int	c_sim;
{
	Cache	*bc[2*NLEVELS];
	char	*save = s;
	int	i, idu, nbc = 0;
	int	bsz = -1, bbsz = -1, ssz;
	int	nsets = 1, repl = -1;
	int	wbt = 0, wa = 0, wi = 0;
	addr_t	sz, t;

	c->c_idu = s[1] + 'A' - 'a';
	c->c_sim = c_sim;
/******************** danalis comm out start ***********
	fprintf(stderr,"******************* c_idu = %d , sto simulation %d \n",c->c_idu,c->c_sim); 
********************* danalis comm out end ************/
	sz = getsize (s+2, &s);
	if (sz <= 0) 
	{
		if (*s)
			shade_fatal ("%s: bad size (%u)", save, sz);
		/*
		 * null cache
		 */
		c->c_nlines = 0;
		c->c_bsize = -1;
		c->c_bbsize = -1;
		c->c_ssize = -1;
		c->c_nsets = 0;
		c->c_replace = C_RANDOM;
		c->c_wthru = 1;
		c->c_walloc = 0;
		c->c_winv = 0;
		c->c_isbkinv = 0;
		c->c_bkinv = 0;
		c->c_tags = 0;
		c->c_valid = 0;
		c->c_dirty = 0;
		return;
	}

	while (*s)
		switch (*s) 
		{
		case 'b':
			if (0 > (bsz = ilog2 (t = getsize (s+1, &s))))
				shade_fatal ("%s: block size (%u) not power of 2",
				    save, t);
			if (*s != ',') 
			{
				bbsz = bsz;
				break;
			}
			if (0 > (bbsz = ilog2 (t = getsize (s+1, &s))))
				shade_fatal ("%s: subblock size (%u) not power of 2",
				    save, t);
			if (bsz < bbsz)
				shade_fatal ("%s: block size (%u) < subblock size (%u)", 
				    save, 1 << bsz, 1 << bbsz);
			break;
		case 'r':
			if (!strncmp (s+1, "lru", 3)) 
			{
				repl = C_LRU;
				s += 4;
			}
			else 
			if (!strncmp (s+1, "random", 6)) 
			{
				repl = C_RANDOM;
				s += 7;
			}
			else 
			if (!strncmp (s+1, "viking", 6)) 
			{
				repl = C_VIKING;
				s += 7;
			}
			else	
				shade_fatal ("%s: bad set replacement", save);
			break;
		case 's':
			nsets = strtol (s+1, &s, 10);
			if (nsets <= 0)
				shade_fatal ("%s: bad set associativity", save);
			break;
		case 'w':
			switch (*++s)
			{
				case 'a':
					wa = *s;
					break;
				case 'b':
				case 't':
					wbt = *s;
					break;
				case 'i':
					wi = 1;
					break;
				default:
					shade_fatal ("%s: w%c invalid", save, *s);
			}
			s++;
			break;
		case 'I':
			idu = s[1];
			i = strtol (s+2, &s, 10);

			if (i <= 0 || i > nlevels ||
			    idu != 'i' && idu != 'd' && idu != 'u' ||
			    (idu == 'u') ^ (icaches[i-1].c_idu == 'U') ||
			    nbc >= 2 * NLEVELS)
				shade_fatal ("%s: bad cache inclusion", save);

			if (idu == 'd')
			{
				if (c_sim == 0)
					bc[nbc++] = dcaches + i - 1;
				else
					bc[nbc++] = fdcache + offset + i -1;
			}
			else
			{
				if (c_sim == 0)
				{
					bc[nbc++] = icaches + i - 1;	/* i or u */
				}
				else
					bc[nbc++] = ficache + offset + i-1;
			}
			break;
		default:
			shade_fatal ("%s: bad cache option (%s)", save, s);
		}

	if (bsz < 0)
		shade_fatal ("%s: bad/missing block size", save);
	/*
	 * insure that a 32 bit word (mask_t) suffices
	 * to hold all of the subblock valid bits.
	 */
	if (bsz - bbsz > 5)
		shade_fatal ("%s: subblock size (%d) << block size (%d)",
		    save, 1 << bbsz, 1 << bsz);

#if 0	/* accesscache divides and conquers */
	/*
	 * insure that doublewords fit in blocks/subblocks so
	 * e.g. ldd's don't have to be treated as 2 ld's.
	 */
	if (bsz <= 2)
		shade_fatal ("%s: blocksize (%d) too small", save, 1 << bsz);
	if (bbsz <= 2)
		shade_fatal ("%s: subblock size (%d) too small",
		    save, 1 << bbsz);
#endif

	if (sz % nsets)
		shade_fatal ("%s: cache size (%s) not multiple of assoc (%d)",
		    save, putsize (sz), nsets);
	if (0 > (ssz = ilog2 (t = sz / nsets)))
		shade_fatal ("%s: set size (%s) not power of 2",
		    save, putsize (t));
	if (ssz < bsz)
		shade_fatal ("%s: set size (%s) < block size (%d)", 
		    save, putsize (t), 1 << bsz);

	if (!wbt)
		wbt = wa ? 'b' : 't';
	if (wbt == 'b')
		wa = 'a';	/* write back implies write allocate */

	c->c_bsize = bsz;
	c->c_bbsize = bbsz;
	c->c_nsets = nsets;
	c->c_ssize = ssz;
	c->c_nlines = 1 << (c->c_ssize - c->c_bsize);
	c->c_walloc = wa == 'a' ? 1 : 0;
	c->c_wthru = wbt == 't' ? 1 : 0;

	c->c_tags = alloc (c->c_nsets, addr_t *);
	c->c_valid = alloc (c->c_nsets, mask_t *);
	if (!c->c_wthru)
		c->c_dirty = alloc (c->c_nsets, mask_t *);
	c->c_used = alloc (c->c_nsets, char *);

	for (i = 0; i < c->c_nsets; i++) 
	{
		c->c_tags[i] = alloc (c->c_nlines, addr_t);
		c->c_valid[i] = alloc (c->c_nlines, mask_t);
		if (!c->c_wthru)
			c->c_dirty[i] = alloc (c->c_nlines, mask_t);
		c->c_used[i] = alloc (c->c_nlines, char);
	}

	if (c->c_nsets > 1) 
	{
		c->c_replace = repl < 0 ? C_RANDOM : repl;
		/*
		 * mru bit mask (1 bit per set) must fit in mask_t.
		 */
		if (c->c_replace == C_VIKING && c->c_nsets > 32)
			shade_fatal ("%s: set associativity too high for viking",
			    save);
		newcachesets (c);
	}
	else 
		if (repl >= 0)
			shade_fatal ("%s: set replacement not allowed for direct mapped cache",
		    	save);
		
	c->c_isbkinv = 0;

	if (nbc <= 0)
		c->c_bkinv = 0;
	else 
	{
		c->c_bkinv = alloc (1 + nbc, Cache *);

		for (i = 0; i < nbc; i++) 
		{
			/* BUG: should check intervening caches too */
			if (!bc[i]->c_wthru)
				shade_fatal ("%s: %c%d not write thru", save,
				    bc[i]->c_idu, bc[i]->c_level);
			if (bc[i]->c_bbsize < 0)
				shade_fatal ("%s: can't include null cache (%c%d)",
				    save, bc[i]->c_idu, bc[i]->c_level);
				    
			c->c_bkinv[i] = bc[i];
			c->c_bkinv[i]->c_isbkinv = 1;
		}
	}

	if (wi)
		if (c->c_wthru && !c->c_walloc && c->c_nsets == 1 && 
		    nlevels == 0)
			c->c_winv = 1;
		else	
			shade_fatal ("%s: wi not allowed for this cache", save);
}


static void
newcachesets (c)
Cache	*c;
{
	int	s;

	switch (c->c_replace) 
	{
		case C_LRU:
			c->c_lru = 1 + alloc (c->c_nsets + 1, uint32 *);
			for (s = -1; s < c->c_nsets; s++)
				c->c_lru[s] = alloc (c->c_nlines, uint32);
			break;
		case C_RANDOM:
			break;
		case C_VIKING:
			c->c_mru = alloc (c->c_nlines, uint32);
			break;
	}
}


static int
replacecacheset (c, X)
Cache	*c;
addr_t	X;
{
	int	s, s2;
	mask_t	mru;

	switch (c->c_replace) 
	{
		case C_LRU:
			for (s2 = 0, s = 1; s < c->c_nsets; s++)
				if (c->c_lru[s][X] < c->c_lru[s2][X])
					s2 = s;
			return (s2);
		case C_RANDOM:
			if (++c->c_random >= c->c_nsets)
				c->c_random = 0;
			return (c->c_random);
		case C_VIKING:
			mru = c->c_mru[X];
			for (s = c->c_nsets-1; s > 0 && (mru & (1 << s)); s--)
					;
			return (s);
	}
	return (0);
}


static void
resetcache (c)
Cache	*c;
{
	invalidatecache (c);

	if (c->c_nsets > 1)
		resetcachesets (c);
}


static void
resetcaches()
{
	int	i;
	int	j, k;

	for (i = 0; i < nlevels; i++) 
	{
		resetcache (icaches + i);
		resetcache (dcaches + i);	/* overkill if unified */
	}

	for (j = 0; j < HOW_MANY; j++)  /* new */
		for (k = 0; k < NLEVELS; k++)
		{
			resetcache(ficache + j+ k);
			resetcache(fdcache + j+ k);
		}
}


static void
resetcachesets (c)
Cache	*c;
{
	addr_t	X;
	int	s;

	switch (c->c_replace) 
	{
		case C_LRU:
			for (s = -1; s < c->c_nsets; s++)
				for (X = 0; X < c->c_nlines; X++)
					c->c_lru[s][X] = 0;
			break;
		case C_RANDOM:
			c->c_random = 0;
			break;
		case C_VIKING:
			for (X = 0; X < c->c_nlines; X++)
				c->c_mru[X] = 0;
			break;
	}
}


cachestats (fp, icaches, dcaches, ni)
FILE	*fp;
Cache	*icaches, *dcaches;
double	ni;
{
	Cache	*ic, *dc;

	fprintf (fp, "\n\n");

	cachestatsc (fp, (Cache *) 0);

	for (ic=icaches, dc=dcaches; ic&&dc; ic=ic->c_next, dc=dc->c_next) 
	{
		cachestatsc (fp, ic);
		if (dc->c_idu == 'D')
			cachestatsc (fp, dc);
	}

	if (ni <= 0)
		return;

	for (ic=icaches, dc=dcaches; ic&&dc; ic=ic->c_next, dc=dc->c_next)
		cachestatsl (fp, ic, dc, ni);
}


cachestatsc (fp, c)
FILE	*fp;
Cache	*c;
{
	Cache	**cc;
	char	wp[32];

	if (!c) 
	{
		fprintf (fp, "%-5s%8s%8s%8s%8s  %-7s  %-18s  %s\n\n",
		    "cache", "size", "block", "subblk", "assoc",
		    "replace", "write policy", "includes");
		return;
	}

	fprintf (fp, "%c%-4d", c->c_idu, c->c_level);

	if (c->c_ssize < 0) 
	{
		fprintf (fp, "%8s\n", "0   ");
		return;
	}

	fprintf (fp, "%8s", putsize ((addr_t) c->c_nsets << c->c_ssize));

	fprintf (fp, "%6u B", 1 << c->c_bsize);

	if (c->c_bsize != c->c_bbsize)
		fprintf (fp, "%6u B", 1 << c->c_bbsize);
	else	
		fprintf (fp, "%8s", "N/A");

	if (c->c_nsets <= 1)
		fprintf (fp, "%8s", "direct");
	else	
		fprintf (fp, "%4d-way", c->c_nsets);

	if (c->c_nsets > 1)
		if (c->c_replace == C_LRU)
			fprintf (fp, "  %-7s", "lru");
		else 
		if (c->c_replace == C_RANDOM)
			fprintf (fp, "  %-7s", "random");
		else 
		if (c->c_replace == C_VIKING)
			fprintf (fp, "  %-7s", "viking");
		else	fprintf (fp, "  %-7s", "N/A");
	else	
		fprintf (fp, "  %-7s", "N/A");

	sprintf (wp, "%s,%salloc%s",
	    c->c_wthru ? "thru" : "back",
	    c->c_walloc ? "" : "no",
	    c->c_wthru && !c->c_walloc ? (c->c_winv ? ",inv" : ",noinv") : "");
	fprintf (fp, "  %-18s  ", wp);

	if (cc = c->c_bkinv)
		for (; *cc; cc++)
			fprintf (fp, "%c%d%s", (*cc)->c_idu, (*cc)->c_level,
			    cc[1] ? "," : "\n");
	else	
		fprintf (fp, "N/A\n");
}


cachestatsl (fp, ic, dc, ni)
FILE	*fp;
Cache	*ic, *dc;
double	ni;
{
	double	r, ref, iref, dref, idrwref[2][2], m, miss, idrwmiss[2][2];
	int	id, rw, sz;

	/*
	 * for null caches, all references miss.
	 */
	if (ic->c_bbsize < 0)
		for (sz = 0; sz <= MAXSIZE; sz++)
			for (rw = 0; rw < 2; rw++)
				ic->c_nmiss[sz][rw] = ic->c_nref[sz][rw];
	if (dc->c_bbsize < 0)
		for (sz = 0; sz <= MAXSIZE; sz++)
			for (rw = 0; rw < 2; rw++)
				dc->c_nmiss[sz][rw] = dc->c_nref[sz][rw];

	iref = dref = ref = miss = 0;
	for (id = 0; id < 2; id++)
		for (rw = 0; rw < 2; rw++)
			idrwref[id][rw] = idrwmiss[id][rw] = 0;

	for (sz = 0; sz <= MAXSIZE; sz++)
		for (rw = 0; rw < 2; rw++) 
		{
			idrwref[0][rw] += ic->c_nref[sz][rw];
			idrwref[1][rw] += dc->c_nref[sz][rw];
			idrwmiss[0][rw] += ic->c_nmiss[sz][rw];
			idrwmiss[1][rw] += dc->c_nmiss[sz][rw];
		}

	for (rw = 0; rw < 2; rw++) 
	{
		iref += idrwref[0][rw];
		dref += idrwref[1][rw];
		for (id = 0; id < 2; id++)
			miss += idrwmiss[id][rw];
	}
	fprintf(fp,"\n\n misses are %lf\n",miss);
	ref = iref + dref;
	fprintf(fp,"\n\n references are %lf\n",ref);

	fprintf (fp, "\n\n%s Level:\n", ordinal (ic->c_level));
	fprintf (fp, "%15s%11s%11s\n\n",
	    "#", "%instrs", "%I+Drefs");

#define doita(x,y)	cachestatsla (fp, (x), ni, ref), \
			fprintf (fp, "  %s", (y))
#define doitb(a,b,c)	fprintf (fp, "  (%.3f%%%s)\n", \
			    (b) ? percent ((a), (b)) : 100., (c))

	doita (ref, "I+D references\n");
	doita (miss, "I+D misses\n");
	fprintf (fp, "\n");

	doita (iref, "I references\n");
	doita (m = idrwmiss[0][0] + idrwmiss[0][1], "I misses");
	doitb (m, iref, "");
	fprintf (fp, "\n");

	doita (dref, "D references\n");
	doita (m = idrwmiss[1][0] + idrwmiss[1][1], "D misses");
	doitb (m, dref, "");
	fprintf (fp, "\n");

	doita (r = idrwref[1][0], "D reads");
	doitb (r, dref, " D refs");
	doita (m = idrwmiss[1][0], "D read misses");
	doitb (m, r, "");
	fprintf (fp, "\n");

	doita (r = idrwref[1][1], "D writes");
	doitb (r, dref, " D refs");
	doita (m = idrwmiss[1][1], "D write misses");
	doitb (m, r, "");
	if (!dc->c_wthru)
		doita (dc->c_nrw, "D read mod writes\n");
	if (dc->c_winv) 
	{
		doita (dc->c_nwinv,"D write invalidates");
		doitb (dc->c_nwinv, idrwref[1][1], " D writes");
	}
	fprintf (fp, "\n");

	if (!ic->c_wthru)    /* I miss causes data write back in I+D cache */
		doita (ic->c_nwback, "write backs due to I misses\n");
	if (!dc->c_wthru)
		doita (dc->c_nwback, "write backs due to D misses\n");
	if (!ic->c_wthru || !dc->c_wthru)
		doita (ic->c_nwback + dc->c_nwback, "write backs\n\n");

	if (ic->c_isbkinv)
		doita (ic->c_nbkinved, "I back invalidated subblocks\n");
	if (dc->c_isbkinv)
		doita (dc->c_nbkinved, "D back invalidated subblocks\n");
	if (ic->c_bkinv)
		doita (ic->c_nbkinving, "I victimized subblocks\n");
	if (dc->c_bkinv)
		doita (dc->c_nbkinving, "D victimized subblocks\n");
	if (ic->c_bkinv || ic->c_isbkinv ||
	    dc->c_bkinv || dc->c_isbkinv)
		fprintf (fp, "\n");


	for (sz = 0; sz <= MAXSIZE; sz++)
		for (rw = 0; rw < 2; rw++) 
		{
			static char *rws[2] = {"read", "write"};

			r = dc->c_nref[sz][rw];
			m = dc->c_nmiss[sz][rw];

			/*
			 * for uniform output, always print all counts
			 * for 1-8 byte references at 1st level.
			 * otherwise only print nonzero counts.
			 * 2nd+ levels should only have 1 data size used,
			 * except when null caches are present.
			 */
			if (!r && !m && (sz > 3 || dc->c_level > 1))
				continue;

			cachestatsla (fp, r, ni, ref);
			fprintf (fp, "  %s D %ss  (%.3f%% D %ss)\n", 
			    putsize ((addr_t) 1 << sz), rws[rw],
			    percent (r, idrwref[1][rw]), rws[rw]);

			cachestatsla (fp, m, ni, ref);
			fprintf (fp, "  %s D %s misses  (%.3f%%)\n", 
			    putsize ((addr_t) 1 << sz), rws[rw],
			    r ? percent (m, r) : (m ? 100. : 0.));
		}

	fprintf (fp, "\n");

	cacheusestats (fp, ic);
	if (dc->c_idu == 'D')
		cacheusestats (fp, dc);
}


cachestatsla (fp, x, ni, ref)
FILE	*fp;
double	x, ni, ref;
{
	fprintf (fp, "%15.0f%10.4f%%", x, percent (x, ni));

#define percent0(x,y)	if (y) \
				fprintf (fp, "%10.4f%%", percent (x, y)); \
			else	fprintf (fp, "%11s", "")

	percent0 (x, ref);
}


cacheusestats (fp, c)
FILE	*fp;
Cache	*c;
{
	int	n, nused = 0, s, X;

	if (c->c_nsets <= 0 || c->c_nlines <= 0 || !c->c_used)
		return;

	for (s = 0; s < c->c_nsets; s++)
		for (X = 0; X < c->c_nlines; X++)
			if (c->c_used[s][X])
				nused++;

	n = c->c_nsets * c->c_nlines;

	fprintf (fp, "%15d  %c%d blocks used (%.4f%% blocks)\n",
	    nused, c->c_idu, c->c_level, percent (nused, n));
	fprintf (fp, "%15d  %c%d blocks unused (%.4f%% blocks)\n",
	    n-nused, c->c_idu, c->c_level, percent (n-nused, n));
}


static char *
domalloc (n)
unsigned int	n;
{
	char	*malloc(), *x = malloc (n);

	if (!x)
		shade_fatal ("can't malloc %u bytes", n);
	return (x);
}


static addr_t
getsize (s, ss)
	char	*s, **ss;
{
	addr_t	u;

	u = strtol (s, &s, 10);

	for (; *s; s++)
		if (*s == 'K')
			u <<= 10;
		else if (*s == 'M')
			u <<= 20;
		else if (*s == 'G')
			u <<= 30;
		else	
			break;

	if (ss)
		*ss = s;
	return (u);
}


static int
ilog2 (u)
addr_t	u;
{
	int	i;

	if (u == 0)
		return (-1);

	for (i = 0; (u & 1) == 0; i++, u >>= 1);

	return (u == 1 ? i : -1);
}


init_memsize()
{
	int	ih, sz;

	for (ih = 0; ih < NIHASH; ih++) 
	{
		for (sz = 0; ihash_memsize[ih] > (1<<sz) && sz < MAXSIZE; sz++);
		ihash_memsize[ih] = sz;
	}
}


static char *
ordinal (card)
int	card;
{
	static char	buf[32], *suf[] = { "th", "st", "nd", "rd" };
	int		n = card % 10;

	if (card >= 11 && card <= 13 || n > 3)
		n = 0;
	sprintf (buf, "%d%s", card, suf[n]);
	return (buf);
}


static char *
putsize (u)
addr_t	u;
{
	static char	buf[16];
	char		*s = "B KBMBGBTB";

	for (; s[2] && (u & 0x3ff) == 0; s += 2, u >>= 10);
	sprintf (buf, "%u %2.2s", u, s);
	return (buf);
}

void
PrintUsefulResults()
{
	int	i;

	if (abel_file)
	{
		fprintf(abel_file,"L1_I_Cache_miss = %d ,\n L1_D_Cache_miss = %d , \nL2_Cache_miss = %d \n",L1_I_Cache_miss, L1_D_Cache_miss, L2_Cache_miss);
		fprintf(abel_file,"L1_I_Total_References = %d , \nL1_D_Total_References = %d , \nL2_Total_References = %d \n", L1_I_Total_References, L1_D_Total_References, L2_Total_References);

		for (i = 0; i < HOW_MANY; i++)
		{
			fprintf(abel_file,"array_L1_I_Cache_miss[%d] = %d \n",i+1,array_L1_I_Cache_miss[i]);
			fprintf(abel_file,"array_L1_D_Cache_miss[%d] = %d \n",i+1,array_L1_D_Cache_miss[i]);
			fprintf(abel_file,"array_L2_Cache_miss[%d] = %d \n",i+1,array_L2_Cache_miss[i]);

			fprintf(abel_file,"array_L1_I_Total_References[%d] = %d \n",i+1,array_L1_I_Total_References[i]);
			fprintf(abel_file,"array_L1_D_Total_References[%d] = %d \n",i+1,array_L1_D_Total_References[i]);
			fprintf(abel_file,"array_L2_Total_References[%d] = %d \n",i+1,array_L2_Total_References[i]);
		}
	}
}

void
init_my_statistic_arrays()
{
	int	i;

	for (i = 0; i < HOW_MANY; i++)
	{
		array_L1_I_Cache_miss[i] = 0;
		array_L1_D_Cache_miss[i] = 0;
		array_L2_Cache_miss[i] = 0;

		array_L1_I_Total_References[i] = 0;
		array_L1_D_Total_References[i] = 0;
		array_L2_Total_References[i] = 0;

		array_ex_L1_D_Cache_miss[i] = 0;
		array_ex_L1_I_Cache_miss[i] = 0;
		array_ex_L2_Cache_miss[i] = 0;

		array_ex_L1_D_Total_References[i] = 0;
		array_ex_L1_I_Total_References[i] = 0;
		array_ex_L2_Total_References[i] = 0;
	}

	for (i = 0; i < HOW_MANY + 1; i ++)
	{
		array_ex_misses[i] = 0;
		array_ex_hits[i] = 0;
		array_ex_InstrMisses[i] = 0;
		array_ex_InstrHits[i] = 0;
	}
	
	for (i = 0; i < AT_MORE; i++)
		TOTAL_COUNTER[i] = 0;

	REPEATS = 0;
	return;
}
